#  Copyright (c) 2020 Maksim Penkov
#  SPDX-License-Identifier: Apache-2.0

import enum
import inspect
import typing


class BindingScope(enum.Enum):
    SINGLETON = 'SINGLETON'
    UNIQUE = 'UNIQUE'


class FactoryType(enum.Enum):
    SIMPLE = 'SIMPLE'
    CONTEXT = 'CONTEXT'


class ArgFactory:
    def __init__(self, factory, factory_type):
        self.factory = factory
        self.factory_type = factory_type


class ArgInstance:
    def __init__(self, instance):
        self.instance = instance


class ScopedInitializer(typing.NamedTuple):
    scope: BindingScope
    initializer: typing.Any


class DependencyContext:
    def __init__(self):
        self.names = {}
        self.classes = {}

    def _target(self, to):
        assert to is not None

        if isinstance(to, str):
            return self.names
        elif inspect.isclass(to):
            return self.classes

        raise ValueError()

    def bind(self, what, to=None, scope=None):
        to = to or _deduce_to(what)

        target = self._target(to)

        if scope is None:
            scope = _default_scope(what)

        if inspect.isclass(what):
            initializer = ArgFactory(
                factory=what,
                factory_type=FactoryType.SIMPLE,
            )
        else:
            initializer = ArgInstance(
                instance=what,
            )

        target[to] = ScopedInitializer(
            scope=scope,
            initializer=initializer,
        )

    def bind_factory(self, factory, to=None, factory_type=None, scope=None):
        to = to or _deduce_to(factory)

        assert factory_type is not None

        scope = scope or BindingScope.UNIQUE
        target = self._target(to)

        target[to] = ScopedInitializer(
            scope=scope,
            initializer=ArgFactory(
                factory=factory,
                factory_type=factory_type,
            )
        )


def _deduce_to(what):
    ret = inspect.signature(what).return_annotation

    if ret == inspect.Signature.empty:
        return None

    return ret


def _default_scope(of):
    return BindingScope.SINGLETON



#  Copyright (c) 2020 Maksim Penkov
#  SPDX-License-Identifier: Apache-2.0

import contextlib
import copy
import functools
import inspect

from spydi.context import (DependencyContext, BindingScope, ArgFactory,
                           ArgInstance, FactoryType, ScopedInitializer)
from spydi.instance_holder import SingletonHolder, UniqueHolder


class DependencyFactory:
    def __init__(self, ctx: DependencyContext):
        self._ctx = ctx

        self._names = {
            name: self._make_holder(relation)
            for name, relation in ctx.names.items()
        }

        self._classes = {
            name: self._make_holder(relation)
            for name, relation in ctx.classes.items()
        }

    def _make_holder(self, relation):
        scope = relation.scope
        initializer = relation.initializer

        if isinstance(initializer, ArgFactory):
            factory = self._factory(initializer)
        elif isinstance(initializer, ArgInstance):
            factory = functools.partial(
                _instance_factory,
                instance=initializer.instance,
            )
        else:
            raise ValueError()

        return _scoped_factory(scope, factory)

    def _factory(self, initializer: ArgFactory):
        simple_factory = functools.partial(
                self.create,
                whatever=initializer.factory,
            )

        def context_factory():
            assert self._stack
            ctx_manager = simple_factory()

            if isinstance(ctx_manager, contextlib.AbstractAsyncContextManager):
                return self._stack.enter_async_context(ctx_manager)
            return self._stack.enter_context(ctx_manager)

        if initializer.factory_type == FactoryType.SIMPLE:
            return simple_factory
        elif initializer.factory_type == FactoryType.CONTEXT:
            return context_factory

        raise ValueError()

    def obtain(self, whatever):
        if whatever not in self._classes:
            self._classes[whatever] = self._make_holder(
                ScopedInitializer(
                    scope=BindingScope.SINGLETON,
                    initializer=ArgFactory(
                        factory=whatever,
                        factory_type=FactoryType.SIMPLE,
                    ),
                ),
            )

        return self._classes[whatever].obtain()

    def create(self, whatever):
        sign = inspect.signature(whatever)

        bound_params = {}

        for name, param in sign.parameters.items():
            name = param.name
            default = param.default
            annotation = param.annotation

            value = inspect.Parameter.empty

            if name in self._names:
                value = self._names[name].obtain()

            if (annotation != inspect.Parameter.empty and
                    value == inspect.Parameter.empty and
                    not _is_primitive(annotation)):
                value = self.obtain(annotation)

            if (default != inspect.Parameter.empty and
                    value == inspect.Parameter.empty):
                value = default

            assert value != inspect.Parameter.empty

            bound_params[name] = value

        def maker(bound_params_dict):
            ba = sign.bind(**bound_params_dict)
            return whatever(*ba.args, **ba.kwargs)

        async def async_maker(bound_params_dict):
            bound_params_awaited = {
                key: (await obj) if inspect.isawaitable(obj) else obj
                for key, obj in bound_params_dict.items()
            }
            bound_params_dict = bound_params_awaited

            result = maker(bound_params_dict)
            if inspect.isawaitable(result):
                return await result
            return result

        if _has_awaitables(bound_params.values()):
            result_maker = async_maker
        else:
            result_maker = maker

        return result_maker(bound_params)

    @contextlib.contextmanager
    def create_ctx(self, whatever):
        with contextlib.ExitStack() as self._stack:
            yield self.create(whatever)

    @contextlib.asynccontextmanager
    async def create_ctx_async(self, whatever):
        async with contextlib.AsyncExitStack() as self._stack:
            yield await self.create(whatever)


def _scoped_factory(scope, factory):
    if scope == BindingScope.SINGLETON:
        return SingletonHolder(factory)
    elif scope == BindingScope.UNIQUE:
        return UniqueHolder(factory)

    raise ValueError()


def _instance_factory(instance):
    return copy.copy(instance)


def _is_primitive(cls):
    return isinstance(cls, (int, str, float, bool))


def _has_awaitables(iterable):
    for testable in iterable:
        if inspect.isawaitable(testable):
            return True

    return False



#  Copyright (c) 2020 Maksim Penkov
#  SPDX-License-Identifier: Apache-2.0

import asyncio
import inspect


class SingletonHolder:
    def __init__(self, factory):
        self._factory = factory
        self._instance = None

    def obtain(self):
        if self._instance is None:
            self._instance = _ensure_awaitable_is_task(
                self._factory(),
            )

        return self._instance


class UniqueHolder:
    def __init__(self, factory):
        self._factory = factory

    def obtain(self):
        return _ensure_awaitable_is_task(
            self._factory(),
        )


def _ensure_awaitable_is_task(obj):
    if inspect.isawaitable(obj) and not isinstance(
            obj,
            asyncio.Task,
    ):
        return asyncio.create_task(obj)

    return obj



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name='spy-di',
    version='0.1.0',
    author='Maksim Penkov',
    author_email='me@madmax.im',
    description='A Simple PYthon Dependency Injection Library',
    keywords='dependency-injection di',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/madmax_inc/spydi',
    packages=['spydi'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
